! WHIZARD 1.13 Aug 06 2001
! 
! (C) 1999-2001 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module decay_forests

  use kinds, only: default !NODEP!
  use limits, only: MAX_EXTERNAL, MAX_CHANNELS, BUFFER_SIZE, DEFAULT_FILENAME
  use mpi90, only: mpi90_broadcast !NODEP!
  use particles
  use decay_trees, decay_tree_TC => TC
  use mappings
  use file_utils !NODEP!
  use parser
  use cascades, only: make_phase_space_file, phase_space_switches
  use diagnostics, only: msg_buffer
  use diagnostics, only: msg_fatal, msg_error, msg_message
  use parameters, only: parameter_set !NODEP!

  implicit none
  private

  integer, parameter, public :: TC = decay_tree_TC

  type, public :: decay_forest
     integer :: n_externals, n_internals, n_in, n_out
     integer :: n_channels_max, n_channels, n_groves
     integer :: n_dim, n_dim0, n_dim1, n_dim2
     integer(kind=TC) :: n_branches_tot, n_branches_decay
     real(kind=default), dimension(:), pointer :: mass_sum
     type(decay_tree), dimension(:), pointer :: tree
     type(mapping), dimension(:,:), pointer :: map
     type(decay_grove), dimension(:), pointer :: grove
  end type decay_forest
  type, public :: decay_grove
     integer :: first, last
  end type decay_grove

  public :: create, destroy
  public :: write
  public :: forest_fill_tmp_file, forest_append_tmp_file
  public :: forest_count
  public :: forest_read
  public :: add
  public :: map
  public :: mpi90_broadcast
  public :: forest_set_strfun
  public :: forest_prepare

  interface create
     module procedure forest_create
  end interface
  interface destroy
     module procedure forest_destroy
  end interface
  interface write
     module procedure forest_write_unit
  end interface
  interface add
     module procedure forest_add_tree
  end interface
  interface map
     module procedure forest_set_mapping
  end interface
  interface mpi90_broadcast
     module procedure forest_broadcast
  end interface

contains

  subroutine forest_create(f, n_in, n_out, n_channels_max)
    type(decay_forest), intent(out) :: f
    integer, intent(in) :: n_in, n_out, n_channels_max
    !allocate(f)
    f%n_in = n_in
    f%n_out = n_out
    f%n_externals = n_in + n_out
    f%n_channels_max  = n_channels_max
    f%n_channels  = 0
    f%n_groves = 0
    f%n_internals = f%n_externals-3
    f%n_branches_tot = 2**f%n_externals-1
    f%n_branches_decay = 2**(n_out)-1
    f%n_dim0  = 0
    f%n_dim1  = n_out - 2
    f%n_dim2  = 2*n_out - 2
    f%n_dim = f%n_dim0 + f%n_dim1 + f%n_dim2
    allocate(f%mass_sum(f%n_branches_tot))
    allocate(f%tree(f%n_channels_max))
    allocate(f%map(f%n_branches_tot, f%n_channels_max))
    f%map = mapping_null
    allocate(f%grove(f%n_channels_max))
  end subroutine forest_create

  subroutine forest_destroy(f)
    type(decay_forest), intent(inout) :: f
    integer :: ch
    deallocate(f%grove)
    deallocate(f%map)
    do ch=1, f%n_channels
       call destroy(f%tree(ch))
    end do
    deallocate(f%tree)
    deallocate(f%mass_sum)
    !deallocate(f)
  end subroutine forest_destroy

  subroutine forest_write_unit(u,f)
    integer, intent(in) :: u
    type(decay_forest), intent(in) :: f
    integer(kind=TC) :: k
    integer :: ch, igrove
    write(u,'(1x,A,1x,I3)') 'Overall dimension:    ', f%n_dim
    write(u,'(1x,A,1x,I3)') 'Phase space dimension:', f%n_dim1+f%n_dim2
    write(u,'(1x,A)') 'Phase space channels:'
    igrove = 1
    do ch=1, f%n_channels
       if (igrove <= f%n_groves) then
          if (ch == f%grove(igrove)%first) &
               & write(u, '(1x,A,I3)') 'Decay grove #', igrove
       end if
       write(u,'(1x,A,I3)') 'Decay tree #', ch
       call write(u,f%tree(ch))
       do k=1, f%n_branches_tot
          call write(u, f%map(k,ch), k)
       end do
       if (igrove <= f%n_groves) then
          if (ch == f%grove(igrove)%last) then
             write(u, '(1x,A)') 'End of grove'
             igrove = igrove+1
          end if
       end if
       write(u,*)
    end do
  end subroutine forest_write_unit

  subroutine forest_fill_tmp_file (filename, unit)
    character(len=*), intent(in) :: filename
    integer, intent(in) :: unit
    call copy_file (trim(filename)//".phs", unit)
    if (trim(filename) /= DEFAULT_FILENAME)  &
         & call copy_file (DEFAULT_FILENAME//".phs", unit)
    call copy_file (trim(filename)//".phx", unit)
  end subroutine forest_fill_tmp_file

  subroutine forest_append_tmp_file (unit, filename)
    integer, intent(in) :: unit
    character(len=*), intent(in) :: filename
    logical :: exist, ok
    integer :: u_out
    u_out = free_unit ()
    inquire (file=filename//".phx", exist=exist)
    if (exist) then
       open (u_out, file=filename//".phx", &
            & action="write", status="old", position="append")
    else
       open (u_out, file=filename//".phx", action="write")
    end if
    rewind (unit)
    call copy_unit (unit, u_out, ok)
    close (u_out)
    if (ok)  call msg_message &
         & (" Wrote phase space configurations to file " &
         &   //filename//".phx")
  end subroutine forest_append_tmp_file

  subroutine copy_file (file, unit)
    character(len=*), intent(in) :: file
    integer, intent(in) :: unit
    integer :: u_in
    logical :: exist
    inquire (file=file, exist=exist)
    if (exist) then
       call msg_message &
            & (" Reading phase space configurations from file "//file)
       u_in = free_unit ()
       open (u_in, file=file, action="read")
       call copy_unit (u_in, unit)
       close (u_in)
    end if
  end subroutine copy_file

  subroutine copy_unit (u_in, u_out, ok)
    integer, intent(in) :: u_in, u_out
    logical, intent(out), optional :: ok
    character(len=BUFFER_SIZE) :: buffer
    integer :: iostat
    if (present(ok))  ok = .false.
    READLINE: do
       read (unit=u_in, fmt="(A)", iostat=iostat) buffer
       if (iostat /= 0)  exit READLINE
       write (unit=u_out, fmt="(A)")  trim(buffer)
       if (present(ok))  ok = .true.
    end do READLINE
  end subroutine copy_unit

  subroutine forest_count &
       & (unit_old, unit_new, filename, process_id, code, par, &
       &  generate, phs, n_channels, new_phase_space)
    integer, intent(in) :: unit_old, unit_new
    character(len=*), intent(in) :: filename, process_id
    integer, dimension(:,:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    logical, intent(in) :: generate
    type(phase_space_switches), intent(in) :: phs
    integer, intent(out) :: n_channels
    logical, intent(out) :: new_phase_space
    rewind (unit_old)
    call forest_count_unit (unit_old, process_id, n_channels)
    if (n_channels == 0) then
       if (generate) then
          call make_phase_space_file &
               & (unit_new, filename, process_id, code, par, n_channels, phs)
          new_phase_space = .true.
       else 
          call msg_message (" No phase space configuration found for process " &
               &            // process_id)
          call msg_message (" If you want me to generate a phase space configuration,")
          call msg_message (" set generate_phase_space = T in the input file.")
          call msg_fatal (" Phase space configuration failed.")
       end if
    else
       write (msg_buffer, "(1x,I5,1x,A)") n_channels, &
            & "phase space channels found for process "//trim(process_id)
       call msg_message
       new_phase_space = .false.
    end if
  end subroutine forest_count

  subroutine forest_count_unit (unit, process_id, n_channels)
    integer, intent(in) :: unit
    character(len=*), intent(in) :: process_id
    integer, intent(out) :: n_channels
    integer :: iostat
    integer :: ch
    logical :: ok
    character(len=BUFFER_SIZE) :: string
    n_channels = 0
    FIND_PROCESS: do
       call find_string(unit, 'process', ok=ok)
       if (.not.ok) exit FIND_PROCESS
       call find_string(unit, process_id, &
            & reject=(/ 'process  ', 'grove    ', 'tree     ', &
            &           'cut      ', 'helicity ', 'flavor   ', &
            &           'histogram', 'and      ' /), &
            & ok=ok)
       if (ok) exit FIND_PROCESS
    end do FIND_PROCESS
    if (.not.ok) then
!        call msg_warning (" No entry found for process "//trim(process_id)//".")
       call lex_clear
       return
    end if
    SCAN_TREES: do ch=1, MAX_CHANNELS
       call find_string(unit, 'tree', reject=(/'process'/), ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_TREES
       n_channels = n_channels+1
       SCAN_MAPPINGS: do
          call find_string(unit, 'map', here=.true., ok=ok, &
               &          iostat=iostat)
          if (iostat < 0) exit SCAN_TREES
          if (.not.ok) exit SCAN_MAPPINGS
          string = get_string(unit)
          string = get_string(unit)
          string = get_string(unit)
       end do SCAN_MAPPINGS
    end do SCAN_TREES
    call lex_clear
  end subroutine forest_count_unit

  subroutine forest_read (unit, f, process_id)
    integer, intent(in) :: unit
    type(decay_forest), intent(inout) :: f
    character(len=*), intent(in) :: process_id
    integer :: iostat
    integer, dimension(2*MAX_EXTERNAL) :: a
    integer :: i, ch, code, mode
    integer(kind=TC) :: k
    logical :: ok, within_grove
    character(len=BUFFER_SIZE) :: string
    character(len=6) :: name
    type(mapping), dimension(f%n_branches_tot) :: default_map
    default_map = mapping_null
    FIND_PROCESS: do
       call find_string(unit, 'process', ok=ok)
       if (.not.ok) exit FIND_PROCESS
       call find_string(unit, process_id, &
            & reject=(/ 'process  ', 'grove    ', 'tree     ', &
            &           'cut      ', 'helicity ', 'flavor   ', &
            &           'histogram', 'and      ' /), &
            & ok=ok)
       if (ok) exit FIND_PROCESS
    end do FIND_PROCESS
    if (.not.ok)  call msg_fatal &
         & (" Can't happen: Process entry not found in phase space configuration.")
    SCAN_DEFAULT_MAPPINGS: do
       call find_string(unit, 'map', reject=(/'process', 'grove  ', 'tree   '/), &
            &           ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_DEFAULT_MAPPINGS
       k = get_integer(unit)
       string = get_string(unit)
       select case(string(1:1))
       case('s','S');  mode = S_CHANNEL
       case('t','T');  mode = T_CHANNEL
       case('u','U');  mode = U_CHANNEL
       case('o','O');  mode = ON_SHELL
       case('n','-');  mode = NO_MAPPING
       case default
          call msg_message (" map "//trim(string)//" ...")
          call msg_error (" This is not a valid mapping specifier.  I'll ignore this.")
          mode = NO_MAPPING
       end select
       name = get_string(unit)
       code = particle_code(name)
       if (mode/=NO_MAPPING .and. code==UNDEFINED) then
          call msg_message (" mapping for particle "//name//":")
          call msg_error (" This particle is undefined.  I'll ignore the mapping.")
          mode = NO_MAPPING
       end if
       call set(default_map(k), mode=mode, code=code)
    end do SCAN_DEFAULT_MAPPINGS
    within_grove = .false.
    SCAN_TREES: do ch=1, MAX_CHANNELS
       call find_string(unit, 'grove', reject=(/'process', 'tree   '/), &
            &           ok=ok, iostat=iostat)
       if (ok .or. ch > f%n_channels_max) then
          call find_string(unit, 'end', reject=(/'process', 'tree   '/), &
               &           ok=ok, iostat=iostat)
          if (ok .or. ch > f%n_channels_max) then
             within_grove = .false.
          else
             f%n_groves = f%n_groves + 1
             f%grove(f%n_groves)%first = ch
             within_grove = .true.
          end if
       end if
       if (within_grove) f%grove(f%n_groves)%last = ch
       call find_string(unit, 'tree', reject=(/'process'/), ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_TREES
       SET_TREE_CODES: do i=1, size(a)
          a(i) = get_integer(unit, ok=ok)
          if (.not.ok) exit SET_TREE_CODES
       end do SET_TREE_CODES
       call forest_add_tree(f, a(:i-1), default_map)
       SCAN_MAPPINGS: do
          call find_string(unit, 'map', here=.true., ok=ok, &
               &          iostat=iostat)
          if (iostat < 0) exit SCAN_TREES
          if (.not.ok) exit SCAN_MAPPINGS
          k = get_integer(unit)
          string = get_string(unit)
          select case(string(1:1))
          case('s','S');  mode = S_CHANNEL
          case('t','T');  mode = T_CHANNEL
          case('u','U');  mode = U_CHANNEL
          case('o','O');  mode = ON_SHELL
          case('n','-');  mode = NO_MAPPING
          case default
             call msg_message (" map "//trim(string)//" ...")
             call msg_error (" This is not a valid mapping specifier.  I'll ignore this.")
             mode = NO_MAPPING
          end select
          name = get_string(unit)
          code = particle_code(name)
          if (mode/=NO_MAPPING .and. code==UNDEFINED) then
             call msg_message (" mapping for particle "//name//":")
             call msg_error (" This particle is undefined.  I'll ignore the mapping.")
             mode = NO_MAPPING
          end if
          call forest_set_mapping(f, k, mode, code, ch)
       end do SCAN_MAPPINGS
    end do SCAN_TREES
    call lex_clear
  end subroutine forest_read

  subroutine forest_add_tree(f, a, default_map)
    type(decay_forest), intent(inout) :: f
    integer(kind=TC), dimension(:), intent(in) :: a
    type(mapping), dimension(:), intent(in), optional :: default_map
    integer :: i,ch
    ch = f%n_channels+1
    if (ch>f%n_channels_max)  then
       call msg_fatal(" Can't happen: Maximum number of phase space channels exceeded in decay forest.")
    end if
    call create(f%tree(ch), f%n_in, f%n_out)
    call tree_from_array(f%tree(ch), a)
    call flip_t_to_s_channel(f%tree(ch))
    if (present(default_map)) then
       do i=1, size(a)
          f%map(flipped(f%tree(ch),a(i)), ch) = default_map(a(i))
       end do
    end if
    f%n_channels = ch
  end subroutine forest_add_tree

  subroutine forest_set_mapping(f, k, mode, code, channel)
    type(decay_forest), intent(inout) :: f
    integer(kind=TC), intent(in) :: k
    integer, intent(in) :: mode, code
    integer, intent(in), optional :: channel
    integer :: ch
    ch = f%n_channels;  if (present(channel)) ch = channel
    if (ch>f%n_channels) then
       write (msg_buffer, "(1x,A,1x,I5,1x,A,1x,I5)") &
            & "ch =", ch, "n_channels =", f%n_channels
       call msg_message
       call msg_error ("Mapping for absent phase space channel: Ignored.")
    else
       call set(f%map(flipped(f%tree(ch),k),ch), mode=mode, code=code)
    end if
  end subroutine forest_set_mapping

  subroutine forest_broadcast(f, root, domain, error)
    type(decay_forest), intent(inout) :: f
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    integer, dimension(:), allocatable :: buffer
    allocate(buffer(size(transfer(f, buffer))))
    buffer = transfer(f, buffer)
    call mpi90_broadcast(buffer, root, domain, error)
    f = transfer(buffer, f)
    deallocate(buffer)
  end subroutine forest_broadcast

  subroutine forest_set_strfun(f, n_strfun)
    type(decay_forest), intent(inout) :: f
    integer, intent(in) :: n_strfun
    f%n_dim0 = n_strfun
    f%n_dim = f%n_dim0 + f%n_dim1 + f%n_dim2
  end subroutine forest_set_strfun

  subroutine forest_prepare(f, code, par)
    type(decay_forest), intent(inout) :: f
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer(kind=TC) :: k
    integer :: i, ch
    f%mass_sum = 0
    do k=1, f%n_branches_decay
       do i=0, ubound(code,1)-1
          if (btest(k,i)) then
             if (ibclr(k,i)==0) then
                f%mass_sum(k) = particle_mass(code(i+1), par)
             else
                f%mass_sum(k) = f%mass_sum(ibclr(k,i))+f%mass_sum(ibset(0,i))
             end if
             exit
          end if
       end do
       do ch=1, f%n_channels
          call set(f%map(k,ch), par, k)
       end do
    end do
  end subroutine forest_prepare


end module decay_forests
